package ch.zhaw.gartenverwaltung.backgroundtasks.weather;

import ch.zhaw.gartenverwaltung.io.CropList;
import ch.zhaw.gartenverwaltung.io.HardinessZoneNotSetException;
import ch.zhaw.gartenverwaltung.io.PlantList;
import ch.zhaw.gartenverwaltung.io.TaskList;
import ch.zhaw.gartenverwaltung.models.PlantNotFoundException;
import ch.zhaw.gartenverwaltung.types.*;

import java.io.IOException;
import java.time.LocalDate;
import java.util.List;

/**
 * The WeatherGardenTaskPlanner creates Tasks based on weather events and the rain amount from the last days
 *
 */
public class WeatherGradenTaskPlanner {
    private final TaskList taskList;
    private final PlantList plantList;
    private final CropList cropList;
    WeatherService weatherService;
    private final LocalDate dateSevereWeather = LocalDate.of(2022,12,15);

    public WeatherGradenTaskPlanner(TaskList taskList, PlantList plantList, CropList cropList) {
        this.taskList = taskList;
        this.plantList = plantList;
        this.cropList = cropList;
        weatherService = new WeatherService();
    }

    /**
     * Method to refresh watering tasks and severe weather tasks
     * @throws IOException If the database cannot be accessed
     * @throws HardinessZoneNotSetException If the hardiness zone is not available
     * @throws PlantNotFoundException if the plant is not available for the watering task
     */
    public void refreshTasks() throws IOException, HardinessZoneNotSetException, PlantNotFoundException {
        getSevereWeatherEvents();
        getRainAmount();
    }

    private void getSevereWeatherEvents() throws IOException {
        SevereWeather actualWeather = weatherService.causeSevereWeather(1);
        if (SevereWeather.HAIL.equals(actualWeather)) {
            createPreHailTask();
        } else if (SevereWeather.FROST.equals(actualWeather)) {
            createPreFrostTask();
        } else if (SevereWeather.SNOW.equals(actualWeather)) {
            createPreSnowTask();
        }
    }

    private void getRainAmount() throws IOException, HardinessZoneNotSetException, PlantNotFoundException {
        int rainAmount = weatherService.causeRainAmount(3);
        adjustWateringTask(rainAmount);
    }

    /**
     * Method to create a PreHailTask
     * @throws IOException If the database cannot be accessed
     */
    private void createPreHailTask() throws IOException {

        Task preHailTask = new Task("Hail",
                                    "During a summer Thunderstorm it could hail heavily. THe Hail could damage the crops. To prevent damage cover the plants with a strong tarpaulin",
                                    dateSevereWeather.minusDays(1L),dateSevereWeather.plusDays(1L));
        taskList.saveTask(preHailTask);

    }

    /**
     * Method to create a PreFrosttask
     * @throws IOException If the database cannot be accessed
     */
    private void createPreFrostTask() throws IOException {
        Task preFrostTask = new Task("Frost",
                                    "The temperatur falls below zero degrees, cover especially the root with wool",
                                    dateSevereWeather.minusDays(1L),dateSevereWeather.plusDays(1L));
        taskList.saveTask(preFrostTask);
    }

    /**
     * Method to create a PreSnowTask
     * @throws IOException If the database cannot be accessed
     */
    private void createPreSnowTask() throws IOException {
        Task preSnowTask = new Task("Snow",
                                    "The weather brings little snowfall. Cover your crops",
                                    dateSevereWeather.minusDays(1L),dateSevereWeather.plusDays(1L));
        taskList.saveTask(preSnowTask);
    }

    /**
     * Method to adjust the water plant tasks
     * @param rainAmount Amount of rain from the last 7 days
     * @throws IOException If the database cannot be accessed
     * @throws HardinessZoneNotSetException If the hardiness zone is not available
     * @throws PlantNotFoundException if the plant is not available for the watering task
     */
    private void adjustWateringTask(int rainAmount) throws HardinessZoneNotSetException, IOException, PlantNotFoundException {

        for (Crop crop : cropList.getCrops()) {
            Plant plant = plantList.getPlantById(HardinessZone.ZONE_8A,crop.getPlantId()).orElseThrow(PlantNotFoundException::new);

            if(plant.wateringCycle().litersPerSqM() < rainAmount){
                adjustNextExecutionOfWateringTasks(taskList.getTaskList(LocalDate.now(), LocalDate.now().plusDays(7)));
            }
        }

    }

    /**
     * Method to set next execution date of the water plant tasks
     * @param cropTaskList List with tasks from crops
     * @throws IOException If the database cannot be accessed
     * @throws HardinessZoneNotSetException If the hardiness zone is not available
     * @throws PlantNotFoundException if the plant is not available for the watering task
     */
    private void adjustNextExecutionOfWateringTasks(List<Task> cropTaskList){
        for(Task task : cropTaskList){
            if(task.getName().equals("water plant")){
                task.setNextExecution(task.getNextExecution().plusDays(task.getInterval().orElse(1)));
            }
        }
    }
}
